struct Uniforms {
	val_seq_1:         vec4<i32>,
	val_seq_2:         vec4<i32>,
	val_seq_env:       vec2<f32>,
	shader_idx:        u32,
	elapsed_time:      f32,
	scene_time:        f32,
	frame_time:        f32,
	sequencer_ticks:   u32,
	screen_flash:      f32,
	screen_width:      i32,
	screen_height:     i32,
	track_volume:      f32,

	pad1: u32,

	val_mod0:          vec4<f32>,
	val_mod1:          vec4<f32>,
	val_mod2:          vec4<f32>,
	val_mod3:          vec4<f32>,
};


const MAX_WAVEFORM_RENDERING = 64;

struct Foo {
	geom: vec4<f32>,
	play_back_pos: f32,
	start : f32,
	end   : f32,
	volume : f32,
};

struct UniformsWave {
	foo: array<Foo, MAX_WAVEFORM_RENDERING>,
};


@group(0) @binding(0) var sampler_bilinear: sampler;

@group(0) @binding(1) var tex_charset:      texture_2d<f32>;
@group(0) @binding(2) var tex_colormap:     texture_2d<f32>;
@group(0) @binding(3) var tex_text_buffer:  texture_2d<u32>;

@group(0) @binding(4) var<uniform> uniforms: Uniforms;

@group(1) @binding(0) var<uniform> uniforms_wave: UniformsWave;
@group(2) @binding(0) var tex_waveform:     texture_1d<f32>;

struct VSOut
{
	@builtin(position) position: vec4f,
	@location(0) uv: vec2f,
	@location(1) instance_id: u32,
};

@vertex
fn vs_wave(
	@builtin(vertex_index) in_vertex_index: u32,
	@builtin(instance_index) in_instance_id: u32
) -> VSOut
{
	var out: VSOut;

	let g = uniforms_wave.foo[in_instance_id].geom;

	if(in_vertex_index == 0u) {
		out.position = vec4f(g.x, g.y, 0.0, 1.0);
		out.uv       = vec2f(0.0, 0.0);
	} else if(in_vertex_index == 1u) {
		out.position = vec4f(g.z, g.y, 0.0, 1.0);
		out.uv       = vec2f(1.0, 0.0);
	} else if(in_vertex_index == 2u) {
		out.position = vec4f(g.x, g.w, 0.0, 1.0);
		out.uv       = vec2f(0.0, 1.0);
	} else {
		out.position = vec4f(g.z, g.w, 0.0, 1.0);
		out.uv       = vec2f(1.0, 1.0);
	}

	out.instance_id = in_instance_id;

	return out;
}

fn get_waveform(uv: vec2<f32>, w: f32, w_y: f32, instance_id: u32) -> vec4<f32>
{
	let tex_dim = textureDimensions(tex_waveform);

	let pb_pos     = uniforms_wave.foo[instance_id].play_back_pos;
	let smpl_start = uniforms_wave.foo[instance_id].start;
	let smpl_end   = uniforms_wave.foo[instance_id].end;
	let volume     = uniforms_wave.foo[instance_id].volume;

	let color_sample  = textureLoad(tex_colormap, vec2i(1,  0), 0).rgb;
	let color_played  = textureLoad(tex_colormap, vec2i(15, 0), 0).rgb;
	let color_outside = textureLoad(tex_colormap, vec2i(3,  0), 0).rgb;
	
	var wave = textureLoad(tex_waveform, i32(uv.y * f32(tex_dim)), 0).rg;

	{
		let pos = i32((uv.y + w_y) * f32(tex_dim));
		if(pos < i32(tex_dim)) {
			let w = textureLoad(tex_waveform, pos, 0).rg;
			wave = vec2f(min(w.x, wave.x), max(w.y, wave.y));
		}
	}

	{
		let pos = i32((uv.y - w_y) * f32(tex_dim));
		if(pos >= 0) {
			let w = textureLoad(tex_waveform, pos, 0).rg;
			wave = vec2f(min(w.x, wave.x), max(w.y, wave.y));
		}
	}

	let inside = uv.y >= smpl_start && uv.y <= smpl_end;

	let cursor_is_outside = pb_pos < smpl_start || pb_pos > smpl_end;
	let is_cursor = abs(pb_pos - uv.y) < 2.0 * w_y;

	let color = select(
		color_outside,
		select(color_played, color_sample, uv.y > pb_pos),
		inside);

	let w1 = smoothstep(wave.x - w, wave.x + w, uv.x);
	let w2 = smoothstep(wave.y - w, wave.y + w, uv.x);

	var opacity = mix(0.0, mix(1.0, 0.0, w2), w1);

	if(is_cursor && cursor_is_outside) {
		opacity = 1.0 - opacity;
	}

	return vec4f(color, opacity * 0.35 * volume);
}

@fragment
fn fs_wave(in: VSOut) -> @location(0) vec4f
{
	let uv = in.uv;
	let w = dpdxFine(uv.x) * 2.0;
	let w_y = dpdyFine(uv.y) * 0.5;

	return get_waveform(uv, w, w_y, in.instance_id);
}

@fragment
fn fs_wave_hor(in: VSOut) -> @location(0) vec4f
{
	let uv = in.uv.yx;
	let w = dpdyFine(uv.x) * 2.0;
	let w_y = dpdxFine(uv.y) * 0.5;

	return get_waveform(uv, w, w_y, in.instance_id);
}
